/**
 * This is the ClothEngine.ts file for the Meta Spark Cloth Simulation Tools.
 * 
 * This file contains the core classes and methods used to create and run a Cloth Simulation in Meta Spark Studio.
 * 
 * To see the Math objects/wrappers see the math.ts file.
 * 
 * Learn more at https://tomaspietravallo.com/blog/meta-spark-cloth-engine
*/

import { Vector, Quaternion, Object3D, PinnableChain, Engine } from "./math";

import Reactive from "Reactive";
import Scene from "Scene";

/**
 * @description This call extends Object3D and provides custom behavior for bones in a skeleton (see: bindMesh) used with the Cloth Simulation Tools
 *
 * The target bone scene object should be contained within a skeleton/armature. To bind the mesh, supply the bone object, skeleton, and base object.
 *
 * This ensures all rotations can be calculated in world space but are corrected for any local rotation the object might carry from the 3D software it was created in/exported from.
 */
export class Bone extends Object3D {
    constructor(props: { pos?: Vector; vel?: Vector; rot?: Quaternion; pinPosition?: boolean; pinRotation?: boolean }) {
        super(props);
    }

    public bindMesh(bone: SceneObjectBase, skeleton: SceneObjectBase, base: SceneObjectBase) {
        this._mesh = bone;
        if (this._mesh) {
            if (!this.pinPosition || !this.pinRotation) this._updateSignals();

            if (!this.pinPosition) {
                this._mesh.worldTransform.position = this.position.point;
            }

            if (!this.pinRotation) {
                this._mesh.worldTransform.rotation =
                    this.rotation.signal
                        .mul(base.transform.rotation.pinLastValue())
                        .mul(skeleton.transform.rotation.pinLastValue())
                        .mul(this._mesh.transform.rotation.pinLastValue());
            }
        }

        return this;
    }
}

/**
 * @description PinnableChainVertical & PinnableChainHorizontal both provide a way to create a chain of objects whose rotations need to be calculated based on the position of the objects in the chain.
 */
export class PinnableChainVertical extends PinnableChain {
    protected sign: number | ScalarSignal;
    protected parentWorldRotation?: SceneObjectBase;
    constructor(objs: Object3D[], length?: number, drag?: number, parentWorldRotation?: SceneObjectBase) {
        let sign: number | ScalarSignal = 1;
        if (parentWorldRotation) {
            const _s = Reactive.vector(1, 0, 0).rotate(parentWorldRotation.worldTransform.rotation).x.sign();
            sign = _s.abs().lt(0.000001).ifThenElse(1, _s);
        }
        super(objs, length, drag);
        this.sign = sign;
        this.parentWorldRotation = parentWorldRotation;
    }

    solveRotations() {
        if (this.objects.length <= 1) return;

        const pinnedSign = typeof this.sign === "number" ? this.sign : this.sign.pinLastValue();

        const absroll = Vector.UP.copy().abs();

        for (let i = 0; i < this.objects.length; i++) {
            let a = i == 0 ? this.objects[1] : this.objects[i];
            let b = i == 0 ? this.objects[0] : this.objects[i - 1];

            // Copy the rotation of the previous bone
            const roll = b.rotation.copy();
            // Force to one axis
            roll.set([roll.w, 0, roll.y, 0]);
            roll.normalize();

            const fwVec = b.position.copy().sub(a.position).normalized;
            fwVec.z *= pinnedSign;

            const la = Quaternion.fromEulerAngles(
                new Vector(
                    Quaternion.lookAt(fwVec, Vector.BACKWARDS) // create lookAt rotation
                        .toEulerArray(),
                ) // convert to euler representation
                    .mul(Vector.ONE.sub(absroll))
                    .getValues(),
            ) // remove the roll from the lookAt rotation
                .mul(Quaternion.fromEulerAngles([Math.PI / 2, 0, 0])); // correct the fact that upVec is M.Vector.BACKWARDS
            // Vector.BACKWARDS.rotate(Quaternion.fromEulerAngles([Math.PI/2,0,0])) == Vector.UP

            (i == 0 ? b : a).rotation.set(roll.mul(la).getValues());
        }
    }

    /**
     * @description Loads a PinnableChainVertical instance from a JSON object.
     * 
     * To store use `toJSON`
     * 
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    static async fromJSON(json: ReturnType<PinnableChainVertical["toJSON"]>, objects?: { [key: string]: Object3D }) {
        const parentWorldRotation = json.parentWorldRotation ? await Scene.root.findFirst(json.parentWorldRotation) : undefined;
        const p = new PinnableChainVertical([], json.length, json.drag, parentWorldRotation);
        const boneParent = await (await Scene.root.findFirst(json.groups[0][0].o)).getParent();
        const boneGrandparent = await boneParent.getParent();
        let groups = [] as { obj: Object3D; spacing: number | undefined }[][];
        for (let i = 0; i < json.groups.length; i++) {
            let group = [] as { obj: Object3D; spacing: number | undefined }[];
            for (let j = 0; j < json.groups[i].length; j++) {
                const e = json.groups[i][j];
                const obj = objects ? objects[e.o] : new Bone({ pinPosition: e.p });
                if (!objects && e.o) {
                    const mesh = await Scene.root.findFirst(e.o).catch((ex) => {
                        throw new Error(
                            `PinnableChainVertical.fromJSON @ Could not find object: ${e.o}. Indices: [${i}][${j}].\n\n${e}\n\n${ex}`,
                        );
                    });
                    obj.position.set(Vector.fromReactive(mesh.worldTransform.position));
                    obj.rotation.set(Quaternion.fromReactive(mesh.worldTransform.rotation));
                    obj.bindMesh(mesh, boneParent, boneGrandparent);
                }
                group.push({ obj, spacing: e.s || json.length });
            }
            group.forEach((o, i) => {
                o.spacing = o.spacing || group[i + 1]?.obj.position.distance(o.obj.position);
            });
            groups.push(group);
        }
        p.groups = groups;
        p.objects = p.groups.flatMap((g) => g.map((e) => e.obj));
        return p;
    }

    toJSON(): ReturnType<PinnableChain["toJSON"]> & { parentWorldRotation?: string } {
        return Object.assign(super.toJSON(), { parentWorldRotation: this.parentWorldRotation?.name });
    }
}

/**
 * @description PinnableChainVertical & PinnableChainHorizontal both provide a way to create a chain of objects whose rotations need to be calculated based on the position of the objects in the chain.
 */
export class PinnableChainHorizontal extends PinnableChain {
    protected sign: number;
    protected axis: "x" | "y" | "z";
    protected parentWorldRotation?: SceneObjectBase;

    constructor(
        objs: Object3D[],
        length?: number,
        drag?: number,
        parentWorldRotation?: SceneObjectBase,
        axis: "x" | "y" | "z" = "x",
    ) {
        let sign: number = 1;
        if (parentWorldRotation) {
            const normQ = Quaternion.fromReactive(parentWorldRotation.worldTransform.rotation).normalized;
            sign = Math.sign(Vector.LEFT.rotate(normQ)[axis]);
        }
        super(objs, length, drag);
        this.sign = sign;
        this.axis = axis;
        this.parentWorldRotation = parentWorldRotation;
    }

    solveRotations() {
        if (this.objects.length <= 1) return;

        for (let i = 0; i < this.objects.length; i++) {
            let a = i == 0 ? this.objects[1] : this.objects[i];
            let b = i == 0 ? this.objects[0] : this.objects[i - 1];

            (i == 0 ? b : a).rotation.set(
                Quaternion.lookAt(b.position.copy().sub(a.position).normalized, Vector.UP.mul(this.sign))
                    .mul(
                        Quaternion.fromEulerAngles([
                            (-Math.PI / 2) * this.sign + Math.PI / 2,
                            (Math.PI / 2) * this.sign,
                            0,
                        ]),
                    )
                    .getValues(),
            );
        }
    }

    /**
     * @description Loads a PinnableChainHorizontal instance from a JSON object.
     * 
     * To store use `toJSON`
     * 
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    static async fromJSON(json: ReturnType<PinnableChainHorizontal["toJSON"]>, objects?: { [key: string]: Object3D }) {
        const parentWorldRotation = json.parentWorldRotation ? await Scene.root.findFirst(json.parentWorldRotation) : undefined;
        const p = new PinnableChainHorizontal([], json.length, json.drag, parentWorldRotation);
        const boneParent = await (await Scene.root.findFirst(json.groups[0][0].o)).getParent();
        const boneGrandparent = await boneParent.getParent();
        p.sign = 1;
        let groups = [];
        for (let i = 0; i < json.groups.length; i++) {
            let group = [];
            for (let j = 0; j < json.groups[i].length; j++) {
                const e = json.groups[i][j];
                const obj = objects ? objects[e.o] : new Bone({ pinPosition: e.p });
                if (!objects && e.o) {
                    const mesh = await Scene.root.findFirst(e.o).catch((ex) => {
                        throw new Error(
                            `PinnableChainHorizontal.fromJSON @ Could not find object: ${e?.o}. Indices: [${i}][${j}].\n\n${e}\n\n${ex}`,
                        );
                    });
                    obj.position.set(Vector.fromReactive(mesh.worldTransform.position));
                    obj.rotation.set(Quaternion.fromReactive(mesh.worldTransform.rotation).getValues());
                    obj.bindMesh(mesh, boneParent, boneGrandparent);
                }
                group.push({ obj, spacing: e.s || json.length });
            }
            group.forEach((o, i) => {
                o.spacing = o.spacing || group[i + 1]?.obj.position.distance(o.obj.position);
            });
            groups.push(group);
        }
        p.groups = groups;
        p.objects = p.groups.flatMap((g) => g.map((e) => e.obj));
        return p;
    }

    toJSON(): ReturnType<PinnableChain["toJSON"]> & { axis?: "x" | "y" | "z", parentWorldRotation?: string } {
        return Object.assign(super.toJSON(), { parentWorldRotation: this.parentWorldRotation?.name });
    }
}

/**
 * @description ClothEngine is the main class that runs the simulation. It takes in a root object and a skeleton object to determine the bones that will be used in the simulation.
 *
 * To construct a new ClothEngine use the following methods:
 * - fromSceneObject
 * - fromJSON
 * - fromLooseJSON
 *
 * The gravity, elasticity, and wind of the simulation can be altered dynamically to adjust how things look and behave. Note: horizontal elasticity only works on the "HIGH" quality setting.
 *
 * The quality setting can be set to "LOW", "MEDIUM", or "HIGH" to adjust the performance of the simulation. It will only be evaluated ONCE (when constructing the new engine).
 *
 * @see https://tomaspietravallo.com/blog/meta-spark-cloth-engine
 */
export class ClothEngine extends Engine {
    protected verticalChains: PinnableChainVertical[];
    protected horizontalChains: PinnableChainHorizontal[];
    protected skeleton: SceneObjectBase;
    protected root: SceneObjectBase;
    public gravity: Vector = new Vector([0, -9.8, 0]);
    public wind: Vector = new Vector([0, 0, 0]);
    public readonly quality: "LOW" | "MEDIUM" | "HIGH";

    protected constructor(props: {
        verticalChains: PinnableChainVertical[];
        horizontalChains: PinnableChainHorizontal[];
        quality: "LOW" | "MEDIUM" | "HIGH";
        root: SceneObjectBase;
        skeleton: SceneObjectBase;
    }) {
        super();
        if (props.quality != "LOW" && props.quality != "MEDIUM" && props.quality != "HIGH") {
            throw new Error(
                `ClothEngine.constructor @ Invalid quality setting: ${props.quality}.\nLearn more: https://www.tomaspietravallo.com/blog/meta-spark-cloth-engine`,
            );
        }

        this.verticalChains = props.verticalChains;
        this.horizontalChains = props.horizontalChains;
        this.quality = props.quality;
        this.root = props.root;
        this.skeleton = props.skeleton;

        this.onRun(({ elapsedTime, deltaMs, frame }) => {
            // Run the simulation every other frame on the LOW quality setting.
            if (this.quality === "LOW") {
                if (frame % 2 == 0) return;
                deltaMs *= 2;
            }

            const mag =
                (this.wind.mag() *
                    (1 + Math.abs(Math.sin(elapsedTime / 733) + Math.abs(Math.cos(elapsedTime / 239))))) /
                deltaMs;
            const forceVector = this.gravity.copy().add(this.wind.copy().mul(mag));

            this.verticalChains.forEach((c) => c.solve(forceVector, deltaMs));

            this.horizontalChains
                .filter((c) => c.objects.some((o) => o.pinPosition))
                .forEach((c) => c.solveRotations());

            // Only compute horizontal constraints on HIGH quality.
            if (this.quality == "HIGH")
                this.horizontalChains.forEach((c) => {
                    c.solve(Vector.ZERO, deltaMs);
                });

            this.verticalChains.forEach((c, i) => c.solveRotations());
            this.updateAllSignals();
        });

        this.verticalChains.forEach((c, i) => c.solveRotations());
    }

    static async createFromScene(props: {
        root: SceneObjectBase;
        defaultLength?: number;
        searchRadius?: number;
        drag?: number;
        quality?: "LOW" | "MEDIUM" | "HIGH";
        pinTopBones?: boolean;
        pinByName?: boolean;
    }) {
        /**
         * Validate inputs and set defaults
         */
        if (!props || !props.root) throw new Error(`Root object not provided @ ClothEngine.createFromScene`);

        if (props.searchRadius == undefined) props.searchRadius = 0.1;
        if (props.drag == undefined) props.drag = 0.9;
        if (typeof props.quality != "string") props.quality = "MEDIUM";
        if (typeof props.pinTopBones !== "boolean") props.pinTopBones = true;
        if (typeof props.pinTopBones !== "boolean") props.pinByName = true;

        const skeleton = await props.root.findFirst("skeleton");

        if (!skeleton) {
            throw new Error(
                `No skeleton found under a skeleton for root object: ${props.root.name} @ ClothEngine.createFromScene\n\nPlease ensure the object you select for Cloth Simulation is properly rigged. The armature needs to both be:\n1) Present in the 3D model\n(2) Bound to the 3D object's mesh.\n\nYou can use the Cloth Simulation Tools Blender addon for help rigging and parenting the armature to the mesh. Learn more: https://tomaspietravallo.com/blog/meta-spark-cloth-engine#tips-for-effective-use`,
            );
        }

        const sceneObjs = (await skeleton.findByPath("*")).filter((a) => a.name.toUpperCase().includes("BONE"));

        await 0;

        if (sceneObjs.length <= 0) {
            throw new Error(
                `No bones found under skeleton (${skeleton.name}) for root object: ${props.root.name} @ ClothEngine.createFromScene`,
            );
        }

        const boneParent = await sceneObjs[0].getParent();
        const boneGrandparent = await boneParent.getParent();

        /**
         * This data is used in the next steps to sort the bones and determine whether bones are part of vertical/horizontal constraints
         */
        let data: {
            mesh: SceneObjectBase;
            pos: Vector;
            localPos: Vector;
            id: number;
            obj: Object3D;
        }[] = [];

        /**
         * Determine the bounding boxes from the bones, and store their positions (in local and world coordinate system)
         */
        let minX!: number, minY!: number, maxX!: number, maxY!: number, minZ!: number, maxZ!: number;

        for (let id = 0; id < sceneObjs.length; id++) {
            const mesh = sceneObjs[id];
            let pos = Vector.fromReactive(mesh.worldTransform.position);
            let localPos = Vector.fromReactive(mesh.transform.position);
            // @ts-ignore
            data[id] = { mesh, pos, localPos, id, obj: undefined };
            minX = minX < pos.x ? minX : pos.x;
            minY = minY < pos.y ? minY : pos.y;
            minZ = minZ < pos.z ? minZ : pos.z;
            maxX = maxX > pos.x ? maxX : pos.x;
            maxY = maxY > pos.y ? maxY : pos.y;
            maxZ = maxZ > pos.z ? maxZ : pos.z;
        }

        /** Is the object wider than it is deep? Construct the chain accordingly */
        let useZOverX = maxZ - minZ > maxX - minX;

        data.sort((a, b) => a.pos.y - b.pos.y);

        let groupsY: (typeof data)[] = [],
            groupsX: (typeof data)[] = [];

        const normAxis = { x: Vector.LEFT, y: Vector.UP, z: Vector.FORWARDS };

        function shouldAddAsHead(p: (typeof data)[0], groups: typeof groupsY, axis: "x" | "y" | "z"): number | true {
            for (let i = 0; i < groups.length; i++) {
                const g = groups[i];
                const tail = groups[i][g.length - 1];
                if (
                    /** Check if the bone is below the tail (this relies on the data being sorted above) */
                    p.pos[axis] > tail.pos[axis] &&
                    /** SQRT2 / 2 makes a 45º cone around the axis */
                    p.pos.copy().sub(tail.pos).normalized.dot(normAxis[axis]) > Math.SQRT2 / 2 &&
                    /** Check if the bone is close enough for it to be considered the same chain */
                    p.localPos.distance(tail.localPos) <= (props.searchRadius as number)
                )
                    return i;
            }
            return true;
        }

        /** Loop over the data and create arrays with the bones for each chain */
        for (let i = 0; i < data.length; i++) {
            const d = data[i];
            const h = shouldAddAsHead(d, groupsY, "y");
            if (h === true) {
                groupsY.push([d]);
            } else {
                groupsY[h].push(d);
            }
        }

        /** Repeat the process on the other axis */
        data.sort((a, b) => (useZOverX ? a.pos.z - b.pos.z : a.pos.x - b.pos.x));

        for (let i = 0; i < data.length; i++) {
            const d = data[i];
            const h = shouldAddAsHead(d, groupsX, useZOverX ? "z" : "x");
            if (h === true) {
                groupsX.push([d]);
            } else {
                groupsX[h].push(d);
            }
        }

        for (let g = 0; g < groupsY.length; g++) {
            const group = groupsY[g];
            group.reverse();
            for (let i = 0; i < group.length; i++) {
                const elem = group[i];
                data[elem.id].obj = new Bone({
                    pos: elem.pos,
                    pinPosition:
                        /** Pin the first bone of each chain (vertically) */
                        (props.pinTopBones && i == 0) ||
                        /** Pin any bones containing PIN in their name (e.g. Bone001_PINNED, Bone_ABC_PIN, etc) */
                        (props.pinByName && !!elem.mesh?.name.toUpperCase().includes("PIN")),
                }).bindMesh(elem.mesh, boneParent, boneGrandparent);
            }
        }

        groupsX.sort((a, b) => a[0].pos.y - b[0].pos.y);

        /** Create all PinnableChains and construct the ClothEngine instance */
        return new ClothEngine({
            verticalChains: groupsY.map(
                (arr) =>
                    new PinnableChainVertical(
                        arr.map((e) => data[e.id].obj),
                        props.defaultLength,
                        props.drag,
                        props.root,
                    ),
            ),
            horizontalChains: groupsX.map(
                (arr) =>
                    new PinnableChainHorizontal(
                        arr.map((e) => data[e.id].obj),
                        props.defaultLength,
                        props.drag,
                        props.root,
                        useZOverX ? "z" : "x",
                    ),
            ),
            quality: props.quality,
            skeleton: skeleton,
            root: props.root,
        });
    }

    public updateHorizontalElasticity(newValue: number) {
        this.horizontalChains.forEach((c) => {
            c.elasticity = newValue;
        });
    }

    public updateVerticalElasticity(newValue: number) {
        this.verticalChains.forEach((c) => {
            c.elasticity = newValue;
        });
    }

    updateAllSignals() {
        this.verticalChains.forEach((c) => {
            c._updateSignals();
        });
    }

    /**
     * @description Outputs an object representation of the Cloth Engine instance and it's target object. This format is used to save and load Cloth Engine instances. This object is meant to be a strict storage format, it is not easy to edit without a deep understanding of the internals.
     *
     * To load the object use `ClothEngine.fromJSON`
     *
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    toJSON() {
        return {
            version: 1,
            mode: "STRICT" as const,
            data_vertical: this.verticalChains.map((c) => c.toJSON()),
            data_horizontal: this.horizontalChains.map((c) => c.toJSON()),
            quality: this.quality,
            root: this.root.name,
            skeleton: this.skeleton.name,
        };
    }

    /**
     * @description Returns an object containing a loose representation of the Cloth Engine instance and it's target object. This format does not store all information (e.g. multiple pinned bones per chain), but is easier to edit manually.
     *
     * If you have a Cloth Engine instance that you want to tweak slightly, then this method may be useful (to load the output use `ClothEngine.fromLooseJSON`)
     *
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    toLooseJSON() {
        return {
            version: 1,
            mode: "LOOSE" as const,
            data_vertical: this.verticalChains.map((c) => c.objects.map((o) => o.mesh.name as string)),
            data_horizontal: this.horizontalChains.map((c) => c.objects.map((o) => o.mesh.name as string)),
            quality: this.quality,
            root: this.root.name,
            skeleton: this.skeleton.name,
        };
    }

    /**
     * @description This method takes in an object (not string) and attempts to cast it into a strict format used to represent Cloth Engine instances and their target objects.
     *
     * This loose format is easier to edit manually, but it's not as robust as the strict format. It's recommended to use the strict format when saving/loading Cloth Engine instances that do not need manual editing.
     *
     * This loose format does not allow for all configurations or features to be saved (e.g. Multiple pinned bones per chain.)
     *
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    static async fromLooseJSON(json: ReturnType<ClothEngine["toLooseJSON"]>) {
        const newJsonInput = {
            version: json.version,
            mode: "STRICT" as const,
            data_vertical: json.data_vertical.map((g) => ({
                length: undefined,
                groups: [g.map((o) => ({ o, p: undefined, s: undefined }))],
                parentWorldRotation: json.root,
            })),
            data_horizontal: json.data_horizontal.map((g) => ({
                length: undefined,
                groups: [g.map((o) => ({ o, p: undefined, s: undefined }))],
                parentWorldRotation: json.root,
            })),
            root: json.root,
            skeleton: json.skeleton,
            quality: json.quality,
        };
        return ClothEngine.fromJSON(newJsonInput);
    }

    /**
     * @description This method takes in an object (not string) and parses it into a Cloth Engine instance. This object should be in the STRICT format used to represent Cloth Engine instances and their target objects.
     *
     * For an easier to edit format, use the `toLooseJSON` & `ClothEngine.fromLooseJSON` methods.
     *
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    static async fromJSON(json: ReturnType<ClothEngine["toJSON"]>) {
        const root = await Scene.root.findFirst(json.root);
        const skeleton = await root.findFirst(json.skeleton);

        const objects = [...json.data_horizontal, ...json.data_vertical]
            .map((d) => d.groups)
            .flat(2)
            .reduce(
                (acc, curr) => ((acc[curr.o] = acc[curr.o] || curr.p || curr.o.toUpperCase().includes("PIN")), acc),
                {} as { [key: string]: Object3D | boolean },
            );

        const meshes = (
            await Promise.all(
                Object.keys(objects).map((objName) =>
                    Scene.root
                        .findFirst(objName)
                        .then((mesh) => ({ name: objName, mesh }))
                        .catch((ex) => {
                            throw new Error(`ClothEngine.fromJSON @ Could not find object: ${objName}. \n\n${ex}`);
                        }),
                ),
            )
        ).reduce((acc, curr) => ((acc[curr.name] = curr.mesh), acc), {} as { [key: string]: SceneObjectBase });

        if (Object.keys(meshes).length <= 0) {
            throw new Error(`ClothEngine.fromJSON @ No objects found in the scene.`);
        }

        const boneParent = await Object.values(meshes)[0].getParent();
        const boneGrandparent = await boneParent.getParent();

        for (const objName in objects) {
            const obj = new Bone({ pinPosition: objects[objName] as boolean });
            const mesh = meshes[objName];
            if (!mesh) throw new Error(`ClothEngine.fromJSON @ Could not find object: ${objName}. ${mesh}.`);

            obj.position.set(Vector.fromReactive(mesh.worldTransform.position));
            obj.oldPosition.set(obj.position);
            obj.rotation.set(Quaternion.fromReactive(mesh.worldTransform.rotation));
            obj.bindMesh(mesh, boneParent, boneGrandparent);

            objects[objName] = obj;
        }

        const verticalChains = await Promise.all(
            json.data_vertical.map((v) => PinnableChainVertical.fromJSON(v, objects as { [key: string]: Object3D })),
        );

        const horizontalChains = await Promise.all(
            json.data_horizontal.map((h) =>
                PinnableChainHorizontal.fromJSON(h, objects as { [key: string]: Object3D }),
            ),
        );

        return new ClothEngine({
            verticalChains,
            horizontalChains,
            quality: json.quality,
            skeleton: skeleton,
            root: root,
        });
    }
}

// 🤖 @ 1723646642: This is a generated file.
// d1cb6b85208898a0db796cabd7b9f288e1545db4
// 7c467c9
// Meta Spark version: 184.0.0.12.245 (Build: 602003560, Commit: 7f1d6866)
// SDK: 181.0
// OS: macOS
