import Scene from 'Scene';
import Reactive from 'Reactive';
import Diagnostics from 'Diagnostics';
import Patches from 'Patches';

import * as M from './math';
import { ClothEngine } from './ClothEngine';
import Blocks from 'Blocks';

(async function () {
    // Read ClothEngine parameters
    const meshName = await Patches.outputs.getStringOrFallback('Mesh name', 'MESH');
    const searchRadius = await Blocks.inputs.getScalar('Bone search radius');
    const defaultLength = await Blocks.inputs.getScalar('Default bone length');
    const quality = await Blocks.inputs.getStringOrFallback('Simulation quality', 'MEDIUM');
    const discovery = await Blocks.inputs.getStringOrFallback('Skeleton discovery', 'AUTO').pinLastValue() as string;

    var engine: ClothEngine;

    if (discovery === 'AUTO') {
        // Create a cloth engine automatically based on the scene hierarchy
        engine = await ClothEngine.createFromScene({
            root: await Scene.root.findFirst(meshName.pinLastValue() || 'MESH'),
            searchRadius: searchRadius.pinLastValue() || 1,
            defaultLength: defaultLength.pinLastValue() || 0,
            quality: quality.pinLastValue() || 'MEDIUM',
        });
    } else {
        // Attempt to parse input as json, fed to fromJSON or fromLooseJSON correspondingly
        try {
            const json = JSON.parse(discovery);
            switch (json.mode) {
                case 'STRICT':
                    engine = await ClothEngine.fromJSON(json);
                    break;
                case 'LOOSE': 
                    engine = await ClothEngine.fromLooseJSON(json);
                    break;
                default:
                    throw new Error(`JSON mode not implemented (${json.mode}).`);
            }
        } catch (error) {
            Diagnostics.log(`Could not parse structure into valid JSON. Skeleton Discovery: ${discovery}.\n\n${error}`);
            throw new Error(`Error occured while parsing JSON "Skeleton discovery" input`);
        }
    }
        
    // Bind inputs to engine
    const gravityPointSignal = await Blocks.inputs.getPoint('Gravity');
    const windDirection = await Blocks.inputs.getPoint('Wind direction');
    const elasticityH = await Blocks.inputs.getScalar('Elasticity (horizontal)');
    const elasticityV = await Blocks.inputs.getScalar('Elasticity (vertical)');

    Reactive.monitorMany({
        x: gravityPointSignal.x,
        y: gravityPointSignal.y,
        z: gravityPointSignal.z,
    }, { fireOnInitialValue: true }).select('newValues').subscribe(({ x, y, z }) => {
        engine.gravity = new M.Vector([x, y, z]);
    });

    Reactive.monitorMany({
        x: windDirection.x,
        y: windDirection.y,
        z: windDirection.z,
    }, { fireOnInitialValue: true }).select('newValues').subscribe(({ x, y, z }) => {
        engine.wind = new M.Vector([x, y, z]);
    });

    elasticityV.monitor({fireOnInitialValue: true}).select('newValue').subscribe(v => engine.updateVerticalElasticity(v));
    elasticityH.monitor({fireOnInitialValue: true}).select('newValue').subscribe(v => engine.updateHorizontalElasticity(v));

    // Set it running
    engine.start();

    // Uncomment the following lines to output json structure
    // Diagnostics.log(engine.toLooseJSON()); // loose json output
    // Diagnostics.log(engine.toJSON()); // strict json output
})();
