/**
 * @description 3D Math library for Meta Spark. Developed by Tomas Pietravallo.
 * Inspired by VOLTS https://github.com/tomaspietravallo/sparkar-volts
 * @see https://tomaspietravallo.com/blog/meta-spark-cloth-engine
*/

import Reactive from "Reactive";
import Scene from "Scene";
import Materials from "Materials";

import Time from "Time";

// Generate a simple uuid, without node APIs https://stackoverflow.com/a/21963136
export function uuid() {
    var u = "",
        i = 0;
    while (i++ < 36) {
        var c = "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx"[i - 1],
            r = (Math.random() * 16) | 0,
            v = c == "x" ? r : (r & 0x3) | 0x8;
        u += c == "-" || c == "4" ? c : v.toString(16);
    }
    return u;
}

/**
 * @description Represents a 3D Vector. Used to represent positions, velocities, forces, etc.
 * 
 * Use the `vector` or `point` properties to get a reactive signal, use fromReactive to feed an xyz signal and get it's pinned representation
 * 
 * Note: this class only stores the vector itself, not the coordinate space it's in.
 * 
 * Inspired by VOLTS https://github.com/tomaspietravallo/sparkar-volts
 * 
 * Learn more: https://tomaspietravallo.com/blog/meta-spark-cloth-engine
 */
export class Vector {
    public static get ONE() {
        return new Vector([+1, +1, +1]);
    }
    public static get ZERO() {
        return new Vector([+0, +0, +0]);
    }
    public static get UP() {
        return new Vector([+0, +1, +0]);
    }
    public static get DOWN() {
        return new Vector([+0, -1, +0]);
    }
    public static get LEFT() {
        return new Vector([+1, +0, +0]);
    }
    public static get RIGHT() {
        return new Vector([-1, +0, +0]);
    }
    public static get FORWARDS() {
        return new Vector([+0, +0, +1]);
    }
    public static get BACKWARDS() {
        return new Vector([+0, +0, -1]);
    }

    public readonly UUID = uuid();

    private values: [number, number, number];
    private signals?: ScalarSignalSource[];
    private _vector?: Vec3Signal;
    private _point?: PointSignal;

    constructor(values?: [number, number, number]) {
        if (values) {
            this.values = [values[0], values[1], values[2]];
        } else {
            this.values = [0, 0, 0];
        }
    }

    static random3D(mag = 1) {
        const angle = Math.random() * Math.PI * 2;
        const vz = Math.random() * 2 - 1;
        const vzBase = Math.sqrt(1 - vz * vz);
        const vx = vzBase * Math.cos(angle);
        const vy = vzBase * Math.sin(angle);
        return new Vector([vx * mag, vy * mag, vz * mag]);
    }

    static fromReactive(xyz: PointSignal | Point3D | Vec3Signal) {
        // @ts-ignore
        return new Vector([xyz.x.pinLastValue(), xyz.y.pinLastValue(), xyz.z.pinLastValue()]);
    }

    public add(other: Vector) {
        this.values[0] += other.values[0];
        this.values[1] += other.values[1];
        this.values[2] += other.values[2];
        return this;
    }

    public sub(other: Vector) {
        this.values[0] -= other.values[0];
        this.values[1] -= other.values[1];
        this.values[2] -= other.values[2];
        return this;
    }

    public mul(other: Vector | number) {
        if (typeof other == "number") {
            this.values[0] *= other;
            this.values[1] *= other;
            this.values[2] *= other;
        } else {
            this.values[0] *= other.values[0];
            this.values[1] *= other.values[1];
            this.values[2] *= other.values[2];
        }
        return this;
    }

    /**
     * Note: Unsafe. Can cause the Vector to turn into a NaN array.
     */
    public div(other: Vector | number) {
        if (typeof other == "number") {
            this.values[0] /= other;
            this.values[1] /= other;
            this.values[2] /= other;
        } else {
            this.values[0] /= other.values[0];
            this.values[1] /= other.values[1];
            this.values[2] /= other.values[2];
        }
        return this;
    }

    public magSq() {
        return this.values[0] ** 2 + this.values[1] ** 2 + this.values[2] ** 2;
    }

    public mag() {
        return this.magSq() ** 0.5;
    }

    public dot(other: Vector) {
        return this.x * other.x + this.y * other.y + this.z * other.z;
    }

    public abs() {
        this.values[0] = Math.abs(this.values[0]);
        this.values[1] = Math.abs(this.values[1]);
        this.values[2] = Math.abs(this.values[2]);
        return this;
    }

    public cross(other: Vector) {
        return new Vector([
            this.values[1] * other.values[2] - this.values[2] * other.values[1],
            this.values[2] * other.values[0] - this.values[0] * other.values[2],
            this.values[0] * other.values[1] - this.values[1] * other.values[0],
        ]);
    }

    public normalize() {
        if (this.magSq() < 0.000001) {
            return this;
        } else {
            return this.div(this.mag());
        }
    }

    public get normalized() {
        return new Vector(this.values).normalize();
    }

    public distance(other: Vector) {
        return other.copy().sub(this).mag();
    }

    public rotate(q: Quaternion) {
        q = q.normalized;
        const x = this.x,
            y = this.y,
            z = this.z;
        const qx = q.x,
            qy = q.y,
            qz = q.z,
            qw = q.w;

        const ix = qw * x + qy * z - qz * y;
        const iy = qw * y + qz * x - qx * z;
        const iz = qw * z + qx * y - qy * x;
        const iw = -qx * x - qy * y - qz * z;

        this.values = [
            ix * qw + iw * -qx + iy * -qz - iz * -qy,
            iy * qw + iw * -qy + iz * -qx - ix * -qz,
            iz * qw + iw * -qz + ix * -qy - iy * -qx,
        ];

        return this;
    }

    public getValues(): [number, number, number] {
        return [...this.values];
    }

    public copy() {
        return new Vector([...this.values]);
    }

    public get x() {
        return this.values[0];
    }
    public get y() {
        return this.values[1];
    }
    public get z() {
        return this.values[2];
    }

    public set x(v: number) {
        this.values[0] = v;
    }
    public set y(v: number) {
        this.values[1] = v;
    }
    public set z(v: number) {
        this.values[2] = v;
    }

    public set(v: [number, number, number] | Vector) {
        if (v instanceof Vector) {
            this.values = [...v.values];
        } else {
            this.values = [...v];
        }
        return this;
    }

    public get vector(): Vec3Signal {
        if (!this.signals)
            this.signals = [
                Reactive.scalarSignalSource("vector-3d-" + this.UUID + "x"),
                Reactive.scalarSignalSource("vector-3d-" + this.UUID + "y"),
                Reactive.scalarSignalSource("vector-3d-" + this.UUID + "z"),
            ];

        this._updateSignals();

        return (this._vector = this._vector
            ? this._vector
            : Reactive.vector(this.signals[0].signal, this.signals[1].signal, this.signals[2].signal));
    }

    public get point(): PointSignal {
        if (!this.signals)
            this.signals = [
                Reactive.scalarSignalSource("point-3d-" + this.UUID + "x"),
                Reactive.scalarSignalSource("point-3d-" + this.UUID + "y"),
                Reactive.scalarSignalSource("point-3d-" + this.UUID + "z"),
            ];

        this._updateSignals();

        return (this._point = this._point
            ? this._point
            : Reactive.point(this.signals[0].signal, this.signals[1].signal, this.signals[2].signal));
    }

    public _updateSignals() {
        this.signals?.forEach((s, i) => s.set(this.values[i]));
    }

    public toString() {
        return `{ ${this.x}, ${this.y}, ${this.z} } @ ${this.UUID.slice(0, 4)}`;
    }
}

/**
 * @description Represents a Quaternion. Used to represent rotations in 3D space.
 * 
 * Use the `signal` property to get a reactive signal, use fromReactive to feed a wxyz signal and get it's pinned representation
 * 
 * Note: This class has been heavily tested and fuzzed to match Meta Spark Studio's engine behavior, but it is not guaranteed to be 100% accurate or to match future versions of the engine.
 * 
 * Inspired by VOLTS https://github.com/tomaspietravallo/sparkar-volts
 * 
 * Learn more: https://tomaspietravallo.com/blog/meta-spark-cloth-engine
 */
export class Quaternion {
    public readonly UUID = uuid();

    private values: [number, number, number, number];
    private signals?: ScalarSignalSource[];
    private _quatSignal: QuaternionSignal;

    public static identity() {
        return new Quaternion();
    }

    static fromEulerAngles(euler: [number, number, number]) {
        const yaw = euler[2];
        const pitch = euler[1];
        const roll = euler[0];
        const cy = Math.cos(yaw * 0.5);
        const sy = Math.sin(yaw * 0.5);
        const cp = Math.cos(pitch * 0.5);
        const sp = Math.sin(pitch * 0.5);
        const cr = Math.cos(roll * 0.5);
        const sr = Math.sin(roll * 0.5);
        return new Quaternion([
            cr * cp * cy + sr * sp * sy,
            sr * cp * cy - cr * sp * sy,
            cr * sp * cy + sr * cp * sy,
            cr * cp * sy - sr * sp * cy,
        ]);
    }

    static createFromAxisAngle(axis: Vector, angle: number): Quaternion {
        const halfAngle = angle * 0.5;
        const s = Math.sin(halfAngle);
        // Meta Spark does not as of v181 perform a normalization here
        // This matches the engine's behavior more closely
        axis = axis.mul(s);
        return new Quaternion([Math.cos(halfAngle), axis.x, axis.y, axis.z]);
    }

    static fromReactive(wxyz: QuaternionSignal | Vec4Signal | Vec4) {
        // @ts-ignore
        return new Quaternion([
            wxyz.w.pinLastValue(),
            wxyz.x.pinLastValue(),
            wxyz.y.pinLastValue(),
            wxyz.z.pinLastValue(),
        ]);
    }

    static lookAt(fwVector: Vector, up: Vector = Vector.UP): Quaternion {
        const fw = fwVector.normalized;
        let lt;
        if (Math.abs(up.dot(fw)) < 0.999) {
            lt = up.cross(fw).normalize();
        } else {
            return Quaternion.fromEulerAngles([-Math.PI / 2, 0, 0]);
        }

        up = fw.cross(lt).normalize();

        const m11 = lt.x;
        const m12 = lt.y;
        const m13 = lt.z;
        const m21 = up.x;
        const m22 = up.y;
        const m23 = up.z;
        const m31 = fw.x;
        const m32 = fw.y;
        const m33 = fw.z;

        const trace = m11 + m22 + m33;
        const q = new Quaternion();

        if (trace > 0.0) {
            let s = Math.sqrt(trace + 1.0);
            q.w = s * 0.5;
            s = 0.5 / s;
            q.x = (m23 - m32) * s; // these could all be replaced by .values[i]
            q.y = (m31 - m13) * s; // or an array that later builds the quaternion
            q.z = (m12 - m21) * s; // probably left to a future Optimized version
        } else {
            if (m11 >= m22 && m11 >= m33) {
                const s = Math.sqrt(1.0 + m11 - m22 - m33);
                const invS = 0.5 / s;
                q.x = 0.5 * s;
                q.y = (m12 + m21) * invS;
                q.z = (m13 + m31) * invS;
                q.w = (m23 - m32) * invS;
            } else if (m22 > m33) {
                const s = Math.sqrt(1.0 + m22 - m11 - m33);
                const invS = 0.5 / s;
                q.x = (m21 + m12) * invS;
                q.y = 0.5 * s;
                q.z = (m32 + m23) * invS;
                q.w = (m31 - m13) * invS;
            } else {
                const s = Math.sqrt(1.0 + m33 - m11 - m22);
                const invS = 0.5 / s;
                q.x = (m31 + m13) * invS;
                q.y = (m32 + m23) * invS;
                q.z = 0.5 * s;
                q.w = (m12 - m21) * invS;
            }
        }
        return q;
    }

    constructor(values?: [number, number, number, number]) {
        if (values) {
            this.values = [values[0], values[1], values[2], values[3]];
        } else {
            this.values = [1, 0, 0, 0];
        }
    }

    public getValues() {
        return [...this.values] as [number, number, number, number];
    }

    public calcNorm(): number {
        return (this.values[0] ** 2 + this.values[1] ** 2 + this.values[2] ** 2 + this.values[3] ** 2) ** 0.5;
    }

    public normalize(): Quaternion {
        const norm = this.calcNorm();
        this.values[0] /= norm;
        this.values[1] /= norm;
        this.values[2] /= norm;
        this.values[3] /= norm;
        return this;
    }

    public inverse(): Quaternion {
        const q = this.calcNorm();
        this.values[0] /= q;
        this.values[1] /= -q;
        this.values[2] /= -q;
        this.values[3] /= -q;
        return this;
    }

    public get normalized() {
        return new Quaternion([...this.values]).normalize();
    }

    public mul(b: Quaternion): Quaternion {
        const w1 = this.values[0];
        const x1 = this.values[1];
        const y1 = this.values[2];
        const z1 = this.values[3];

        const w2 = b.w;
        const x2 = b.x;
        const y2 = b.y;
        const z2 = b.z;

        this.values[0] = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2;
        this.values[1] = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2;
        this.values[2] = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2;
        this.values[3] = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2;
        return this;
    }

    public toEulerArray() {
        let [w, x, y, z] = [this.w, this.x, this.y, this.z];
        const norm = this.calcNorm();
        w /= norm;
        x /= norm;
        y /= norm;
        z /= norm;
        const angles: number[] = [];
        // roll (x-axis rotation)
        const sinr_cosp = 2 * (w * x + y * z);
        const cosr_cosp = 1 - 2 * (x * x + y * y);
        angles[0] = Math.atan2(sinr_cosp, cosr_cosp);
        // pitch (y-axis rotation)
        // this is intentional, matches Meta Spark's behavior
        const sinp = 2 * (this.w * this.y - this.z * this.x);
        if (Math.abs(sinp) >= 1) {
            angles[1] = (Math.PI / 2) * Math.sign(sinp); // use 90 degrees if out of range
        } else {
            angles[1] = Math.asin(sinp);
        }
        // yaw (z-axis rotation)
        const siny_cosp = 2 * (w * z + x * y);
        const cosy_cosp = 1 - 2 * (y * y + z * z);
        angles[2] = Math.atan2(siny_cosp, cosy_cosp);
        return angles as [number, number, number];
    }

    public copy() {
        return new Quaternion([...this.values]);
    }

    public get w() {
        return this.values[0];
    }

    public get x() {
        return this.values[1];
    }

    public get y() {
        return this.values[2];
    }

    public get z() {
        return this.values[3];
    }

    public _updateSignals() {
        this.signals?.forEach((s, i) => s.set(this.values[i]));
    }

    public get signal() {
        if (!this.signals)
            this.signals = [
                Reactive.scalarSignalSource("quat-" + this.UUID + "w"),
                Reactive.scalarSignalSource("quat-" + this.UUID + "x"),
                Reactive.scalarSignalSource("quat-" + this.UUID + "y"),
                Reactive.scalarSignalSource("quat-" + this.UUID + "z"),
            ];

        this._updateSignals();

        return (this._quatSignal = this._quatSignal
            ? this._quatSignal
            : Reactive.quaternion(
                  this.signals[0].signal,
                  this.signals[1].signal,
                  this.signals[2].signal,
                  this.signals[3].signal,
              ));
    }

    public set w(v: number) {
        this.values[0] = v;
    }

    public set x(v: number) {
        this.values[1] = v;
    }

    public set y(v: number) {
        this.values[2] = v;
    }

    public set z(v: number) {
        this.values[3] = v;
    }

    public set(v: [number, number, number, number] | Quaternion) {
        if (v instanceof Quaternion) {
            this.values = [...v.values];
        } else {
            this.values = [...v];
        }
        return this;
    }
}

/**
 * @description Represents a 3D Object (e.g.: mesh, point, etc)
 */
export class Object3D {
    public readonly position: Vector;
    public readonly oldPosition: Vector;
    public readonly rotation: Quaternion;
    public readonly id: number;
    public readonly pinPosition: boolean;
    public readonly pinRotation: boolean;
    public readonly dynamic = false;
    protected _mesh?: SceneObjectBase;
    private static id = 0;

    constructor(
        props: { pos?: Vector; vel?: Vector; rot?: Quaternion; pinPosition?: boolean; pinRotation?: boolean } = {},
    ) {
        this.position = props.pos ? props.pos.copy() : new Vector();
        this.oldPosition = this.position.copy();
        this.rotation = props.rot ?? new Quaternion();
        this.pinPosition = !!props.pinPosition;
        this.pinRotation = !!props.pinRotation;
        this.id = Object3D.id++;
    }

    distance(other: Object3D) {
        return this.position.distance(other.position);
    }

    async useDynamicMeshBody(root?: SceneObjectBase) {
        this.dynamic = true;
        const body = await Scene.create("Plane", {
            transform: Reactive.transform(this.position.point, Reactive.scale(1, 1, 1), this.rotation.signal),
        });
        await (root ?? Scene.root).addChild(body);
        this._mesh = body;
    }

    public _updateSignals({ pos, rot } = { pos: true, rot: true }) {
        if (pos) {
            if (this.pinPosition) {
                this.position.set(Vector.fromReactive(this._mesh.worldTransform.position).getValues());
            } else this.position._updateSignals();
        }
        if (rot) {
            if (this.pinRotation) {
                this.rotation.set(Quaternion.fromReactive(this._mesh.worldTransform.rotation).getValues());
            } else this.rotation._updateSignals();
        }
    }

    public get mesh() {
        return this._mesh;
    }

    /**
     * @param mesh Mesh to be bound
     * @param args Only used by classes that extend Object3D. See: Bone.bindMesh
     */
    public bindMesh(mesh: SceneObjectBase, ...args: SceneObjectBase[]) {
        this._mesh = mesh;
        if (this._mesh) {
            if (!this.pinPosition || !this.pinRotation) this._updateSignals();

            if (!this.pinPosition) {
                this._mesh.worldTransform.position = this.position.point;
            }

            if (!this.pinRotation) {
                this._mesh.worldTransform.rotation = this.rotation.signal;
            }
        }

        return this;
    }

    public get velocity() {
        // @ts-ignore
        return this.position
            .copy()
            .sub(this.oldPosition)
            .div(this.deltaMs ?? 0.033333);
    }

    static async createDebugMaterial(hue?: number): Promise<MaterialBase> {
        function hsv2rgb(h: number, s: number, v: number): [number, number, number] {
            h *= 360;
            /* istanbul ignore next */
            const f = (n: number, k = (n + h / 60) % 6) => v - v * s * Math.max(Math.min(k, 4 - k, 1), 0);
            return [f(5), f(3), f(1)];
        }

        if (hue === undefined) hue = 0;
        return Materials.create("DefaultMaterial", {
            opacity: 1.0,
            blendMode: "ALPHA",
            doubleSided: true,
        }).then((m) => {
            return m.setTextureSlot("DIFFUSE", Reactive.pack4(...hsv2rgb(hue, 1, 1), 1) as any), m;
        });
    }
}

export interface Constraint {
    solve(): void;
}

/**
 * Position constraint
 * @param length Fixed distance at which the two positions should remain at all times
 */
export class Link implements Constraint {
    private a: Object3D;
    private b: Object3D;
    private length: number;
    private static readonly epsilon = 0.0001;

    constructor(a: Object3D, b: Object3D, length: number) {
        this.a = a;
        this.b = b;
        this.length = length;
    }

    solve() {
        const diff = this.b.position.copy().sub(this.a.position);
        const mag = diff.mag();
        const off = Math.abs(mag - this.length) > Link.epsilon;
        if (off) {
            const f = +!this.a.pinPosition + +!this.b.pinPosition;
            if (f == 0) return;
            diff.normalize()
                .mul(mag - this.length)
                .div(f);
            !this.a.pinPosition && this.a.position.add(diff);
            !this.b.pinPosition && this.b.position.sub(diff);
        }
    }

    toJSON() {
        return {
            a: this.a.mesh?.name,
            b: this.b.mesh?.name,
            length: this.length,
        };
    }
}

/**
 * @description Represents a chain of objects in which there may be more than 1 pinned object.
 * 
 * This is class introduces custom and novel behavior which are not physically accurate, but was tested to fall within a 10% error margin (sacrificing quality for performance).
 * 
 * @author Tomas Pietravallo
 * @see https://tomaspietravallo.com/blog/meta-spark-cloth-engine
 */
export class PinnableChain implements Constraint {
    public objects: Object3D[];
    public length?: number;
    public groups: { obj: Object3D; spacing: number | null }[][];
    public elasticity: number = 1;

    constructor(objs: Object3D[], length?: number, drag = 1) {
        this.objects = objs;
        this.length = length;
        this.elasticity = drag ?? 1;

        let lastGroupIdx: null | number = null;
        let groups: { obj: Object3D; spacing: number | null }[][] = [[]];
        let gC = 0;
        let storeSpacing = !length;

        // Grouping objects
        for (let i = 0; i < this.objects.length; i++) {
            const spacing =
                !storeSpacing || i == this.objects.length - 1
                    ? null
                    : this.objects[i].position.distance(this.objects[i + 1].position);
            groups[gC].push({ obj: this.objects[i], spacing });

            if (i == 0 && this.objects[0].pinPosition) lastGroupIdx = 0;
            if (i !== 0 && this.objects[i].pinPosition) {
                if (lastGroupIdx == null) {
                    gC += 1;
                    groups[gC] = [{ obj: this.objects[i], spacing }];
                } else if (lastGroupIdx == i - 1) {
                    const e = groups[gC].pop()!;
                    groups[++gC] = [e];
                } else {
                    const count = Math.ceil((i - gC) / 2);
                    // splice could potentially be optimized
                    groups[gC + 1] = groups[gC].splice(groups[gC].length - count);

                    // unnecessary
                    // groups[gC].length = count;

                    gC++;

                    if (i !== this.objects.length - 1) {
                        groups[++gC] = [{ obj: this.objects[i], spacing }];
                    }
                }
                lastGroupIdx = i;
            }
        }

        this.groups = groups;
    }

    solve(accForce: Vector = new Vector([0, -9.8, 0]), deltaMs = 33.3333) {
        deltaMs /= 1000;

        const fdt = accForce.copy().mul(deltaMs ** 2);

        for (let gid = 0; gid < this.groups.length; gid++) {
            const group = this.groups[gid];

            // @todo Refactor repeated code
            // Solve forwards
            for (let i = 0; i < group.length; i++) {
                const ed = group[i];
                const e = ed.obj;
                if (e.pinPosition) continue;

                const pos = e.position.copy();
                const vdt = e.position.copy().sub(e.oldPosition).mul(this.elasticity);

                if ((i == 0 && gid > 0) || (i == group.length - 1 && gid < this.groups.length - 1)) {
                    let a: Object3D,
                        b = e;
                    if (i == 0) {
                        a = this.groups[gid - 1][0].obj;
                    } else {
                        a = this.groups[gid + 1][0].obj;
                    }
                    const f = +!a.pinPosition + +!b.pinPosition; // f is never zero. Logically unreachable. `if (e.pinPosition) continue;` check above
                    const diff = b.position.copy().sub(a.position);
                    const mag = diff.mag();
                    diff.normalize()
                        .mul(mag - (ed.spacing ?? this.length))
                        .div(f);
                    !a.pinPosition && a.position.add(diff);
                    !b.pinPosition && b.position.sub(diff);
                }

                e.position.add(vdt);
                e.position.add(fdt);

                if (i >= 1) {
                    const prev = group[i - 1];
                    const norm = e.position
                        .copy()
                        .sub(prev.obj.position)
                        .normalize()
                        .mul(prev.spacing ?? this.length);
                    e.position.set(prev.obj.position.copy().add(norm).getValues());
                }

                e.oldPosition.set(pos.getValues());
            }

            // Solve backwards
            for (let i = group.length - 1; i >= 0; i--) {
                const ed = group[i];
                const e = ed.obj;
                if (e.pinPosition) continue;

                // Group boundaries
                if ((i == 0 && gid > 0) || (i == group.length - 1 && gid < this.groups.length - 1)) {
                    let a: Object3D,
                        b = e;
                    if (i == 0) {
                        a = this.groups[gid - 1][0].obj;
                    } else {
                        a = this.groups[gid + 1][0].obj;
                    }

                    const f = +!a.pinPosition + +!b.pinPosition; // f is never zero. Logically unreachable. `if (e.pinPosition) continue;` check above
                    const diff = b.position.copy().sub(a.position);
                    const mag = diff.mag();
                    diff.normalize()
                        .mul(mag - (ed.spacing ?? this.length))
                        .div(f);
                    !a.pinPosition && a.position.add(diff);
                    !b.pinPosition && b.position.sub(diff);
                }

                // Within the group -> iterate normally
                if (i <= group.length - 2) {
                    const prev = group[i + 1].obj;
                    const norm = e.position
                        .copy()
                        .sub(prev.position)
                        .normalize()
                        .mul(ed.spacing ?? this.length);
                    e.position.set(prev.position.copy().add(norm).getValues());
                }
            }
        }
    }

    /**
     * @description Outputs an object representation of the PinnableChain. This format is used to save and load PinnableChain instances. It's not advisable to modify this without some understanding of the internals.
     *
     * To load the object use `fromJSON`
     *
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    toJSON(): { groups: { o: string; p: boolean; s?: number }[][]; length?: number; drag?: number } {
        return {
            groups: this.groups.map((g) => g.map((e) => ({ o: e.obj.mesh?.name as string, p: e.obj.pinPosition, s: e.spacing }))),
            length: this.length,
            drag: this.elasticity,
        };
    }

    /**
     * @description Loads a PinnableChain instance from a JSON object.
     * 
     * To store use `toJSON`
     * 
     * @see https://gist.github.com/tomaspietravallo/e35d5ebe035933798c9944f7ca1eb202
     */
    static async fromJSON(json: ReturnType<PinnableChain["toJSON"]>) {
        const p = new PinnableChain([], json.length, json.drag);
        let groups = [];
        for (let i = 0; i < json.groups.length; i++) {
            let group = [];
            for (let j = 0; j < json.groups[i].length; j++) {
                const e = json.groups[i][j];
                const obj = new Object3D({ pinPosition: e.p });
                if (e.o) {
                    const mesh = await Scene.root.findFirst(e.o).catch((ex) => {
                        throw new Error(
                            `PinnableChain.fromJSON @ Could not find object: ${e?.o}. Indices: [${i}][${j}].\n\n${e}\n\n${ex}`,
                        );
                    });
                    obj.position.set(Vector.fromReactive(mesh.worldTransform.position));
                    obj.rotation.set(Quaternion.fromReactive(mesh.worldTransform.rotation).getValues());
                    obj.bindMesh(mesh);
                }
                group.push({ obj, spacing: e.s });
            }
            groups.push(group);
        }
        p.groups = groups;
        p.objects = p.groups.flatMap((g) => g.map((e) => e.obj));
        return p;
    }

    _updateSignals() {
        this.objects.forEach((o) => {
            o._updateSignals({ pos: true, rot: true });
        });
    }
}

/**
 * @description Basic Chain implementation with a fixed length between objects
 */
export class Chain implements Constraint {
    private objs: Object3D[];
    private length: number;
    private static readonly epsilon = 0.0001;

    constructor(objs: Object3D[], length: number) {
        this.objs = objs;
        this.length = length;
    }

    solve() {
        for (let i = 0; i < this.objs.length; i++) {
            const a = this.objs[i];
            const b = this.objs[i + 1];
            if (!b) return;

            b.rotation.set(
                Quaternion.lookAt(b.position.copy().sub(a.position).normalized, Vector.FORWARDS).getValues(),
            );

            const diff = b.position.copy().sub(a.position);
            const mag = diff.mag();
            const off = Math.abs(mag - this.length) > Chain.epsilon;
            if (off && !b.pinPosition) {
                diff.normalize().mul(this.length);
                b.position.set(diff.add(a.position).getValues());
            }
        }
    }

    toJSON() {
        return {
            objs: this.objs.map((o) => o.mesh?.name),
            length: this.length,
        };
    }
}

/**
 * @description Base class to represent Physics engines (e.g.: Cloth Engine, etc). Provides start() and stop() methods which control a callback loop that runs on every frame.
 */
export class Engine {
    protected subscription?: Subscription;
    protected _onRun?: ((props: { deltaMs: number; elapsedTime: number; frame: number }) => void)[] = [];
    protected frameCount: number = 0;

    constructor() {}

    public start() {
        if (this.subscription) return this.subscription;

        /* @test-code-remove-start */
        this.subscription = (Time.ms as ScalarSignal).monitor().subscribe((val) => {
            // istanbul ignore next
            const diff = val.newValue - (val.oldValue ?? 0);
            this._onRun &&
                this._onRun.forEach((f) => f({ deltaMs: diff, elapsedTime: val.newValue, frame: this.frameCount }));
            this.frameCount++;
        });
        /* @test-code-remove-end */
    }

    public onRun(cb: (props: { deltaMs: number; elapsedTime: number; frame: number }) => void) {
        this._onRun.push(cb);
    }

    public stop() {
        this.subscription?.unsubscribe();
        this.subscription = undefined;
    }
}

/**
 * @description Sample Physics Engine that applies gravity to objects
 */
export class PhysicsEngine extends Engine {
    gravity: number;
    objects: Object3D[];
    constructor(props: { gravity?: number } = {}) {
        super();
        this.gravity = props.gravity ?? -9.8;
        this.objects = [];
        this.onRun(({ elapsedTime, deltaMs, frame }) => {
            this.objects.forEach((o) => {
                this.applyGravityToObject(o, deltaMs);
            });
        });
    }
    addObject(o: Object3D) {
        this.objects.push(o);
    }
    applyGravityToObject(o: Object3D, deltaMs: number) {
        if (!o.pinPosition) {
            deltaMs /= 1000;
            const pos = o.position.copy();
            // @ts-ignore
            o.deltaMs = deltaMs;

            o.position.set(
                o.position
                    .copy()
                    .add(o.position.copy().sub(o.oldPosition))
                    .add(new Vector([0, this.gravity, 0]).copy().mul(deltaMs ** 2))
                    .getValues(),
            );

            o.oldPosition.set(pos.getValues());
        }
    }
}

// 🤖 @ 1723646642: This is a generated file.
// 0677af3babc4a6b65ea27168561a0c65d83914a3
// 7c467c9
// Meta Spark version: 184.0.0.12.245 (Build: 602003560, Commit: 7f1d6866)
// SDK: 181.0
// OS: macOS
